package dev.shadowsoffire.apotheosis.adventure.socket;

import dev.shadowsoffire.apotheosis.Apotheosis;
import net.minecraft.world.entity.Entity;
import dev.shadowsoffire.apotheosis.adventure.affix.AffixHelper;
import java.util.stream.Stream;
import java.util.Collections;
import net.minecraft.world.entity.projectile.AbstractArrow;
import java.util.Iterator;
import net.minecraft.nbt.ListTag;
import java.util.List;
import java.util.Collection;
import com.google.common.collect.ImmutableList;
import net.minecraft.nbt.Tag;
import net.minecraft.core.NonNullList;
import dev.shadowsoffire.apotheosis.adventure.socket.gem.GemInstance;
import dev.shadowsoffire.apotheosis.adventure.loot.LootCategory;
import java.util.Objects;
import java.util.function.ToIntFunction;
import java.util.function.Function;
import dev.shadowsoffire.placebo.util.CachedObject;
import net.minecraft.nbt.CompoundTag;
import net.minecraftforge.eventbus.api.Event;
import net.minecraftforge.common.MinecraftForge;
import dev.shadowsoffire.apotheosis.adventure.event.GetItemSocketsEvent;
import net.minecraft.world.item.ItemStack;
import net.minecraft.resources.ResourceLocation;

public class SocketHelper
{
    public static final ResourceLocation GEMS_CACHED_OBJECT;
    public static final String AFFIX_DATA = "affix_data";
    public static final String GEMS = "gems";
    public static final String SOCKETS = "sockets";
    
    public static int getSockets(final ItemStack stack) {
        final CompoundTag afxData = stack.m_41737_("affix_data");
        final int sockets = (afxData != null) ? afxData.m_128451_("sockets") : 0;
        final GetItemSocketsEvent event = new GetItemSocketsEvent(stack, sockets);
        MinecraftForge.EVENT_BUS.post((Event)event);
        return event.getSockets();
    }
    
    public static void setSockets(final ItemStack stack, final int sockets) {
        stack.m_41698_("affix_data").m_128405_("sockets", sockets);
    }
    
    public static SocketedGems getGems(final ItemStack stack) {
        return (SocketedGems)CachedObject.CachedObjectSource.getOrCreate(stack, SocketHelper.GEMS_CACHED_OBJECT, (Function)SocketHelper::getGemsImpl, (ToIntFunction)SocketHelper::hashSockets);
    }
    
    private static int hashSockets(final ItemStack stack) {
        return Objects.hash(stack.m_41737_("affix_data"), getSockets(stack));
    }
    
    private static SocketedGems getGemsImpl(final ItemStack stack) {
        final int size = getSockets(stack);
        if (size <= 0 || stack.m_41619_()) {
            return SocketedGems.EMPTY;
        }
        final LootCategory cat = LootCategory.forItem(stack);
        if (cat.isNone()) {
            return SocketedGems.EMPTY;
        }
        final List<GemInstance> gems = (List<GemInstance>)NonNullList.m_122780_(size, (Object)GemInstance.EMPTY);
        int i = 0;
        final CompoundTag afxData = stack.m_41737_("affix_data");
        if (afxData != null && afxData.m_128441_("gems")) {
            final ListTag gemData = afxData.m_128437_("gems", 10);
            for (final Tag tag : gemData) {
                final ItemStack gemStack = ItemStack.m_41712_((CompoundTag)tag);
                gemStack.m_41764_(1);
                final GemInstance inst = GemInstance.socketed(stack, gemStack);
                if (inst.isValid()) {
                    gems.set(i++, inst);
                }
                if (i >= size) {
                    break;
                }
            }
        }
        return new SocketedGems(ImmutableList.copyOf((Collection<? extends GemInstance>)gems));
    }
    
    public static void setGems(final ItemStack stack, final SocketedGems gems) {
        final CompoundTag afxData = stack.m_41698_("affix_data");
        final ListTag gemData = new ListTag();
        for (final GemInstance inst : gems) {
            gemData.add((Object)inst.gemStack().m_41739_(new CompoundTag()));
        }
        afxData.m_128365_("gems", (Tag)gemData);
    }
    
    public static boolean hasEmptySockets(final ItemStack stack) {
        return getGems(stack).gems().stream().anyMatch(g -> !g.isValid());
    }
    
    public static int getFirstEmptySocket(final ItemStack stack) {
        final SocketedGems gems = getGems(stack);
        for (int socket = 0; socket < gems.size(); ++socket) {
            if (!gems.get(socket).isValid()) {
                return socket;
            }
        }
        return 0;
    }
    
    private static List<ItemStack> getGems(final AbstractArrow arrow) {
        final CompoundTag afxData = arrow.getPersistentData().m_128469_("affix_data");
        final int sockets = (afxData != null) ? afxData.m_128451_("sockets") : 0;
        if (sockets <= 0) {
            return Collections.emptyList();
        }
        final List<ItemStack> gems = (List<ItemStack>)NonNullList.m_122780_(sockets, (Object)ItemStack.f_41583_);
        int i = 0;
        if (afxData != null && afxData.m_128441_("gems")) {
            final ListTag gemData = afxData.m_128437_("gems", 10);
            for (final Tag tag : gemData) {
                final ItemStack gemStack = ItemStack.m_41712_((CompoundTag)tag);
                gemStack.m_41764_(1);
                if (GemInstance.unsocketed(gemStack).isValidUnsocketed()) {
                    gems.set(i++, gemStack);
                }
                if (i >= sockets) {
                    break;
                }
            }
        }
        return gems;
    }
    
    public static Stream<GemInstance> getGemInstances(final AbstractArrow arrow) {
        final LootCategory cat = AffixHelper.getShooterCategory((Entity)arrow);
        if (cat == null) {
            return Stream.empty();
        }
        return getGems(arrow).stream().map(gemStack -> GemInstance.socketed(cat, gemStack)).filter(GemInstance::isValid);
    }
    
    static {
        GEMS_CACHED_OBJECT = Apotheosis.loc("gems");
    }
}
